/*--------------------------------------------------------------------
  File:        xgraphics.c
  Author:      Nils Schmeisser
  ----------------------------------------------------------------------
  Description: Graphics module for X applications
  ----------------------------------------------------------------------
  Notes:       
  Date:        3.12.1993
  ----------------------------------------------------------------------
*/ 

#define XTFUNCPROTO
#define ANSI


#include <iostream.h>

#include <math.h>
#include <stdio.h>
#include <string.h>
//#include <values.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/X.h>
#include <Xm/DrawingA.h>

#include "xgraph.h"
#include "xapp.h"

typedef void (*XGraphicsRedrawFunction)(Window,int);
typedef XPoint polypoint;

int xDARKBLUE;
int xDARKGREEN;
int xDARKCYAN;
int xDARKRED;
int xDARKMAGENTA;
int xDARKBROWN;
int xBLACK;
int xBLUE;
int xGREEN;
int xCYAN;
int xRED;
int xMAGENTA;
int xBROWN;
int xLIGHTGRAY;
int xDARKGRAY;
int xLIGHTBLUE;
int xLIGHTGREEN;
int xLIGHTCYAN;
int xLIGHTRED;
int xLIGHTMAGENTA;
int xYELLOW;
int xWHITE;

typedef struct GW {
          Window window;
          float ratio;
          int  x,y,width,height;
          int  foreground,background;
          int  penx,peny;
          int  fill_pattern;
          XGraphicsRedrawFunction redraw;
          GW   *next;
        };

static int ColorOfs=40;
static Colormap ColorTable;
static XColor   ColorEntry[256];
static Window root;
static Window default_window;
static int screen;
static Visual *VisualInfo;
static int LMAXCOLORS=256-40;
static Display *GRAPHICS_DISPLAY=NULL;
static GC GRAPHICS_CONTEXT;

static GW *GW_ROOT=NULL;
static GW *default_gw=NULL;
static char fontname[256];
static Font font;
static int fontwidth=6;
static int fontheight=13;

//=== Error messages ==============================================================

static void Error(char *s) {
  cout << "XGRAPH error message: " << s << "\n";
}

//=== Private, handling for Windows information ===================================

static GW *FIND_GW(Window window) {
  GW *p;

  if (GW_ROOT!=NULL) {
      p=GW_ROOT;
      while ((p!=NULL) && (p->window!=window)) p=p->next;
      return p;
  } else
    return NULL;
}

static void NEW_GW(Window window,int x,int y,int w,int h) {
  GW *p;

  if (GW_ROOT==NULL) {
    GW_ROOT=new GW;
    p=GW_ROOT;
  } else {
    p=GW_ROOT;
    while (p->next!=NULL) p=p->next;
    p->next=new GW;
    p=p->next;
  }
  p->window=window;
  p->x=x;
  p->y=y;
  p->width=w;
  p->height=h;
  p->next=NULL;
  p->ratio=0;
}

static void DEL_GW(Window window) {
  GW *p,*pw;

  if (GW_ROOT!=NULL) {
    pw=FIND_GW(window);
    if (pw!=NULL) {
      if (pw==GW_ROOT) {
	if (pw->next!=NULL) 
	  GW_ROOT=pw->next; 
	else 
	  GW_ROOT=NULL;
	delete pw;
      } else {
	p=GW_ROOT;
	while (p->next!=pw) p=p->next;
	p->next=pw->next;
	delete pw;
      }  
    }
  }
}

static void SET_foreground(Window window,int color) {
  FIND_GW(window)->foreground=color;
}

static void SET_background(Window window,int color) {
  FIND_GW(window)->background=color;
}

static void SET_pen(Window window,int x,int y) {
  GW *p;

  p=FIND_GW(window);
  p->penx=x;
  p->peny=y;
}

static int GET_foreground(Window window) {
  return(FIND_GW(window)->foreground);
}

static int GET_background(Window window) {
  return(FIND_GW(window)->background);
}

static void GET_pen(Window window,int &x,int &y) {
  GW *p;

  p=FIND_GW(window);
  x=p->penx;
  y=p->peny;
}

int ConvRGBtoPIX(int red,int green,int blue) {
  int i,c;
  double dist,min=(256.0*256.0*256.0*256.0)*3.0+1.0;
  double dr,dg,db;

  for (i=0;i<256;i++) {
    dr=(double)ColorEntry[i].red-(double)red*256.0;
    dg=(double)ColorEntry[i].green-(double)green*256.0;
    db=(double)ColorEntry[i].blue-(double)blue*256.0;
    dist=dr*dr+dg*dg+db*db;
    if (dist<min) {
      min=dist;
      c=i;
    }
  }
  return(c);
}

static void create_standard_colors() {
  xBLACK       =ConvRGBtoPIX(  0,  0,  0)-ColorOfs;
  xDARKGRAY    =ConvRGBtoPIX( 63, 63, 63)-ColorOfs;
  xDARKBLUE    =ConvRGBtoPIX(  0,  0, 63)-ColorOfs;
  xDARKGREEN   =ConvRGBtoPIX(  0, 63,  0)-ColorOfs;
  xDARKCYAN    =ConvRGBtoPIX(  0, 63, 63)-ColorOfs;
  xDARKRED     =ConvRGBtoPIX( 63,  0,  0)-ColorOfs;
  xDARKMAGENTA =ConvRGBtoPIX( 63,  0, 63)-ColorOfs;
  xDARKBROWN   =ConvRGBtoPIX( 63, 63,  0)-ColorOfs;

  xBLUE        =ConvRGBtoPIX(  0,  0,127)-ColorOfs;
  xGREEN       =ConvRGBtoPIX(  0,127,  0)-ColorOfs;
  xCYAN        =ConvRGBtoPIX(  0,127,127)-ColorOfs;
  xRED         =ConvRGBtoPIX(127,  0,  0)-ColorOfs;
  xMAGENTA     =ConvRGBtoPIX(127,  0,127)-ColorOfs;
  xBROWN       =ConvRGBtoPIX(127,127,  0)-ColorOfs;

  xLIGHTGRAY   =ConvRGBtoPIX(127,127,127)-ColorOfs;
  xLIGHTBLUE   =ConvRGBtoPIX(  0,  0,255)-ColorOfs;
  xLIGHTGREEN  =ConvRGBtoPIX(  0,255,  0)-ColorOfs;
  xLIGHTCYAN   =ConvRGBtoPIX(  0,255,255)-ColorOfs;
  xLIGHTRED    =ConvRGBtoPIX(255,  0,  0)-ColorOfs;
  xLIGHTMAGENTA=ConvRGBtoPIX(255,  0,255)-ColorOfs;
  xYELLOW      =ConvRGBtoPIX(255,255,  0)-ColorOfs;
  xWHITE       =ConvRGBtoPIX(255,255,255)-ColorOfs;
}

//=================================================================================

Display *getgraphicsdisplay() { return(GRAPHICS_DISPLAY); }

void setgraphicsdisplay(Display *dpy) { GRAPHICS_DISPLAY=dpy; }

void setwindow(Window window) { 
  XWindowAttributes Attr;
  GW *p;

  default_window=window;
  p=FIND_GW(default_window);

  if (p!=NULL) {
    XGetWindowAttributes(GRAPHICS_DISPLAY,window,&Attr);
    p->height=Attr.height;
    p->width=Attr.width;  
    p->x=Attr.x;
    p->y=Attr.y;
    XSetForeground(GRAPHICS_DISPLAY,GRAPHICS_CONTEXT,p->foreground+ColorOfs);
    XSetWindowBackground(GRAPHICS_DISPLAY,window,p->background+ColorOfs);
  } else {
    printf("Window %d is not in graphics window list.\n",window);
  }
}

Window openwindow(char *name,int x,int y,int sizex,int sizey,Widget parent,Widget *da) {
  XSetWindowAttributes Attr;
  XEvent event;
  Window window;
  Widget gshell,drawarea;
  Arg args[10];
  int n=0;

  gshell=xCreatePopupShell(name,x,y,sizex,sizey,parent);
  drawarea=xCreateDrawingArea(sizex,sizey,gshell);
  XtUnmanageChild(drawarea);
  xPopup(gshell);
  XFlush(GRAPHICS_DISPLAY);
  window=XtWindow(gshell);

  Attr.event_mask=ExposureMask | ButtonPressMask | ColormapChangeMask | LeaveWindowMask | 
                  EnterWindowMask;
//  Attr.colormap  =ColorTable;
  Attr.background_pixel=0;
  Attr.border_pixel    =1;
  XChangeWindowAttributes(GRAPHICS_DISPLAY,window,
                          CWBackPixel | CWBorderPixel |
                          CWEventMask /*| CWColormap*/,&Attr);
  XFlush(GRAPHICS_DISPLAY); 
  XMapWindow(GRAPHICS_DISPLAY,window);
  XSelectInput(GRAPHICS_DISPLAY,window,
               ButtonPressMask | ButtonReleaseMask | PointerMotionMask | KeyPressMask |
               /*ColormapChangeMask |*/ ExposureMask | LeaveWindowMask | EnterWindowMask);
  XFlush(GRAPHICS_DISPLAY);
  NEW_GW(window,x,y,sizex,sizey);
  default_window=window;
  xRegisterWindow(window);
  if (da!=NULL) *da=drawarea;
//  sleep(1);
  return(window);
}

void cleardevice(Window);

static unsigned int EVENTKEYCODE=0;

void RedrawXGraphicsWindow(Widget win,XtPointer,XEvent *event,Boolean*) {
  XWindowAttributes Attr;
  Window rootw,d;
  GW *p;
  int reason;
  int x,y;
  unsigned int xa,ya,height,width,depth,border;
  XEvent cevent;
  Boolean resize=False;
  XExposeEvent        expose;

  d=XtWindow(win);
  p=FIND_GW(d);
  if (p!=NULL) {
    XFlush(GRAPHICS_DISPLAY);
    XGetGeometry(GRAPHICS_DISPLAY,d,&rootw,&x,&y,&width,&height,&border,&depth);
    switch (event->type) {
        case ButtonPress: reason=ButtonPress;
                          break;
           case KeyPress: reason=KeyPress;
	                  EVENTKEYCODE=event->xkey.keycode;
                          break;
             case Expose: expose   =event->xexpose;
                          reason=Expose;
                          while (expose.count!=0) {
                            XNextEvent(GRAPHICS_DISPLAY,event);
                            expose=event->xexpose;
                          }
                          
                          // looking ahead if next event in the queue is a resize
                          
                          if (XPending(GRAPHICS_DISPLAY)>0) {
                            XPeekEvent(GRAPHICS_DISPLAY,&cevent);                                
                            if ((cevent.type==ResizeRequest) || (cevent.type==Expose)) {
                              reason=ResizeRequest;
                              XNextEvent(GRAPHICS_DISPLAY,&cevent);
                              if (XPending(GRAPHICS_DISPLAY)>0) {
                                XPeekEvent(GRAPHICS_DISPLAY,&cevent);
                                while ((cevent.type==ResizeRequest) || (cevent.type==Expose)) {
                                  XNextEvent(GRAPHICS_DISPLAY,&cevent);
                                  if (XPending(GRAPHICS_DISPLAY)>0) 
                                    XPeekEvent(GRAPHICS_DISPLAY,&cevent);
                                  else
                                    cevent.type=0;
                                }
                              }
                            }
                          }
                          break;
    }

    if ((height!=p->height) || (width!=p->width)) {
      reason=ResizeRequest;
      p->height=height;
      p->width=width;
    } else if ((reason==ResizeRequest) && ((height==p->height) || (width==p->width))) {
      reason=Expose;
    }
    xa=p->x;
    ya=p->y;
    p->x=x;
    p->y=y;

    if (reason==ResizeRequest) {
      if (p->ratio>0) {
	if (p->width!=(int)((float)height*p->ratio)) {
	  p->width=(int)((float)height*p->ratio);
	  XResizeWindow(GRAPHICS_DISPLAY,d,p->width,p->height);
	}
      }
      cleardevice(d);
    }
    p->redraw(d,reason);
  }
}

void closewindow(Window window) {
  XEvent event;
  Widget draw;

  if (FIND_GW(window)!=NULL) {
    while (XCheckWindowEvent(GRAPHICS_DISPLAY,window,
                             ExposureMask | ButtonPressMask | ColormapChangeMask | 
                             LeaveWindowMask | EnterWindowMask,&event));
    XFlush(GRAPHICS_DISPLAY);

    DEL_GW(window);
    if (window==default_window) {
      if (GW_ROOT!=NULL) 
	default_window=GW_ROOT->window;
      else 
	default_window=MAXINT;
    }

    draw=XtWindowToWidget(GRAPHICS_DISPLAY,window);
    xPopdown(draw);
  }
}

void setredrawfunction(Window window,XGraphicsRedrawFunction redraw_function) {
  Widget draw;

  draw=XtWindowToWidget(GRAPHICS_DISPLAY,window);
  XtAddEventHandler(draw,
                    ExposureMask | StructureNotifyMask | ButtonPressMask | KeyPressMask,
                    False,RedrawXGraphicsWindow,NULL);
  FIND_GW(window)->redraw=redraw_function;
}

Window openwindow(char *name,int x,int y,int sizex,int sizey,
                  Widget parent,XGraphicsRedrawFunction redraw_function,Widget *da) {

  Window window=openwindow(name,x,y,sizex,sizey,parent,da);
  setredrawfunction(window,redraw_function);
  return window;
}

//---------------------------------------------------------------------------------

void setrgbpalette(int colornum, int red, int green, int blue) {
  ColorEntry[colornum+ColorOfs].pixel  =(unsigned long)(colornum+ColorOfs);
  ColorEntry[colornum+ColorOfs].red    =(unsigned short)(red*256);
  ColorEntry[colornum+ColorOfs].green  =(unsigned short)(green*256);
  ColorEntry[colornum+ColorOfs].blue   =(unsigned short)(blue*256);
  ColorEntry[colornum+ColorOfs].flags  = DoRed | DoGreen | DoBlue;    
  XStoreColor(GRAPHICS_DISPLAY,ColorTable,&ColorEntry[colornum+ColorOfs]);
  XFlush(GRAPHICS_DISPLAY);
}

void getrgbpalette(int colornum, int &red, int &green, int &blue) {
  red  =ColorEntry[colornum+ColorOfs].red/256;
  green=ColorEntry[colornum+ColorOfs].green/256;
  blue =ColorEntry[colornum+ColorOfs].blue/256;
}

void cleardevice(Window window) {
  if (window==0) {
    if (default_window!=MAXINT) XClearWindow(GRAPHICS_DISPLAY,default_window); 
  } else XClearWindow(GRAPHICS_DISPLAY,window);
  XFlush(GRAPHICS_DISPLAY);
}

//---------------------------------------------------------------------------------

void initgraph(Display *display) {
  char driver[40];
  int i;
  XColor q;
  XGCValues GCValues;

// get number of x,y,colors

  if (display==NULL) 
    GRAPHICS_DISPLAY=XOpenDisplay((char*)NULL); 
  else 
    GRAPHICS_DISPLAY=display;
  screen=XDefaultScreen(GRAPHICS_DISPLAY);
  root  =XDefaultRootWindow(GRAPHICS_DISPLAY);
  VisualInfo=XDefaultVisual(GRAPHICS_DISPLAY,screen);
//  LMAXCOLORS=256; // should be replaced later by the obtained value
  ColorTable=XDefaultColormap(GRAPHICS_DISPLAY,screen);
  for (i=0;i<LMAXCOLORS;i++) {
    q.pixel=(unsigned long)i;
    XQueryColor(GRAPHICS_DISPLAY,ColorTable,&q);
    ColorEntry[i]=q;
  }

  #ifndef USE_ILUT
    ColorTable=XCreateColormap(GRAPHICS_DISPLAY,root,VisualInfo,AllocAll);
    for (i=0;i<40;i++) {
      XAllocColor(GRAPHICS_DISPLAY,ColorTable,&ColorEntry[i]);
    }
    for (i=40;i<256;i++) {
      ColorEntry[i].pixel=(unsigned long)i;
      ColorEntry[i].flags=DoRed | DoGreen | DoBlue;
      ColorEntry[i].red  =0;
      ColorEntry[i].green=0;
      ColorEntry[i].blue =0;
      XAllocColor(GRAPHICS_DISPLAY,ColorTable,&ColorEntry[i]);
    }
    XStoreColors(GRAPHICS_DISPLAY,ColorTable,ColorEntry,256);

    XInstallColormap(GRAPHICS_DISPLAY,ColorTable);
  #endif

  create_standard_colors();
  GCValues.function  =GXcopy;
  GCValues.plane_mask=AllPlanes;
  GCValues.foreground=1;
  GCValues.background=0;
  GRAPHICS_CONTEXT=XCreateGC(GRAPHICS_DISPLAY,root,
                             GCFunction | GCPlaneMask | GCForeground | GCBackground,&GCValues);

  XFlush(GRAPHICS_DISPLAY);
}

void closegraph(Display *display) {
  XFreeGC(GRAPHICS_DISPLAY,GRAPHICS_CONTEXT);
  if (display==NULL) XCloseDisplay(GRAPHICS_DISPLAY);
}

//-------------------------------------------------------------------

int getmaxcolor() { return(LMAXCOLORS); }

int getmaxx(Window window) { 
  if (window==0) {
    if (default_window!=MAXINT)
      return FIND_GW(default_window)->width-1;
    else
      return 0;
  } else return FIND_GW(window)->width-1;
}

int getmaxy(Window window) { 
  if (window==0) {
    if (default_window!=MAXINT)
      return FIND_GW(default_window)->height-1;
    else
      return 0;
  } else return FIND_GW(window)->height-1;
}

void getsize(Window window,int &w,int &h) {
  GW *p;

  p=FIND_GW(window);
  w=p->width;
  h=p->height;
}

void getsize(int &w,int &h) { getsize(default_window,w,h); }

void getposition(Window window,int &x,int &y) {
  XWindowAttributes Attr;
  Widget drawable;
  Arg args[10];
  int n;
  GW *p;
  Window rootwin;
  Position x0,y0;
  unsigned int w,h,b,d;


  drawable=XtWindowToWidget(GRAPHICS_DISPLAY,window);
  n=0;
  XtSetArg(args[n],XmNx,&x0); n++;
  XtSetArg(args[n],XmNy,&y0); n++;
  XtGetValues(drawable,args,n);
  p=FIND_GW(window);
  p->x=(int)x0;
  p->y=(int)y0;
  x=p->x;
  y=p->y;
}

void getposition(int &x,int &y) { getposition(default_window,x,y); }

//-------------------------------------------------------------------

void setcolor(Window window,int color) {
  XSetForeground(GRAPHICS_DISPLAY,GRAPHICS_CONTEXT,color+ColorOfs);
  FIND_GW(window)->foreground=color;
}

void setcolor(int color) { setcolor(default_window,color); }

int getfgcolor(Window window) { 
  if (window==0) {
    if (default_window!=MAXINT)
      return FIND_GW(default_window)->foreground;
    else
      return 0;
  } else return FIND_GW(window)->foreground;
}

void setbkcolor(Window window,int color) {
  XSetWindowBackground(GRAPHICS_DISPLAY,window,color+ColorOfs);
  FIND_GW(window)->background=color;
}

void setbkcolor(int color) { setbkcolor(default_window,color); }

int getbkcolor(Window window) { 
  if (window==0) {
    if (default_window!=MAXINT)
      return FIND_GW(default_window)->background;
    else
      return 0;
  } else 
    return FIND_GW(window)->background;
}

void keepratio(Window window,Boolean keepit) {
  GW *p=FIND_GW(window);

  if (keepit) p->ratio=(float)(p->width)/(float)(p->height);
  else p->ratio=-1;
}

void keepratio(Boolean keepit) { keepratio(default_window,keepit); }

//--- moveto --------------------------------------------------------

void moveto(Window window,int x,int y) { SET_pen(window,x,y); }
void moveto(int x,int y) { SET_pen(default_window,x,y); }

//--- putpixel ------------------------------------------------------

void putpixel(Window window,int x,int y,int color) {
  if (color>-MAXINT) XSetForeground(GRAPHICS_DISPLAY,GRAPHICS_CONTEXT,color+ColorOfs);
    else XSetForeground(GRAPHICS_DISPLAY,GRAPHICS_CONTEXT,FIND_GW(window)->foreground+ColorOfs);
  XDrawPoint(GRAPHICS_DISPLAY,window,GRAPHICS_CONTEXT,x,y);
//  XFlush(GRAPHICS_DISPLAY);
}

void putpixel(int x,int y,int col) { putpixel(default_window,x,y,col); }

//--- putpixel ------------------------------------------------------

int getpixel(Window window,int x,int y) {
  XImage *p;  

  int w=1;
  int h=11;
  p=XGetImage(GRAPHICS_DISPLAY,window,x,y,w,h,0xffffffff,ZPixmap);
  return XGetPixel(p,0,0);
}

int getpixel(int x,int y) { return getpixel(default_window,x,y); }

//-------------------------------------------------------------------

void line(Window window,int x1,int y1,int x2,int y2,int color) {
  if (color>-MAXINT) XSetForeground(GRAPHICS_DISPLAY,GRAPHICS_CONTEXT,color+ColorOfs);
    else XSetForeground(GRAPHICS_DISPLAY,GRAPHICS_CONTEXT,FIND_GW(window)->foreground+ColorOfs);
  XDrawLine(GRAPHICS_DISPLAY,window,GRAPHICS_CONTEXT,x1,y1,x2,y2);	
//  XFlush(GRAPHICS_DISPLAY);
}

void line(int x1,int y1,int x2,int y2,int color) { 
  if (default_window!=MAXINT) 
    line(default_window,x1,y1,x2,y2,color);
  else
    Error("Unkown default window.");
}

void lineto(Window window,int x,int y,int color) {
  GW *p;
  int px,py;

  p=FIND_GW(window);
  px=p->penx;
  py=p->peny;
  if (color>-MAXINT) XSetForeground(GRAPHICS_DISPLAY,GRAPHICS_CONTEXT,color+ColorOfs);
    else XSetForeground(GRAPHICS_DISPLAY,GRAPHICS_CONTEXT,FIND_GW(window)->foreground+ColorOfs);
  XDrawLine(GRAPHICS_DISPLAY,window,GRAPHICS_CONTEXT,px,py,x,y);	
//  XFlush(GRAPHICS_DISPLAY);
  p->penx=x;
  p->peny=y;
}

void lineto(int x,int y,int color) { 
  if (default_window!=MAXINT) 
    lineto(default_window,x,y,color);
  else
    Error("Unkown default window.");
}

void drawpoly(Window window,int num,polypoint *points,int color) {
  XPoint *Points;

  Points=(XPoint*)points;
  if (color>-MAXINT) XSetForeground(GRAPHICS_DISPLAY,GRAPHICS_CONTEXT,color+ColorOfs);
    else XSetForeground(GRAPHICS_DISPLAY,GRAPHICS_CONTEXT,FIND_GW(window)->foreground+ColorOfs);
  XDrawLines(GRAPHICS_DISPLAY,window,GRAPHICS_CONTEXT,Points,num,CoordModeOrigin);
//  XFlush(GRAPHICS_DISPLAY);
}

void drawpoly(int num,polypoint *points,int color) { 
  if (default_window!=MAXINT) 
    drawpoly(default_window,num,points,color); 
  else
    Error("Unkown default window.");
}

void fillpoly(Window window,int num,polypoint *points,int color) {

  if (color>-MAXINT) XSetForeground(GRAPHICS_DISPLAY,GRAPHICS_CONTEXT,color+ColorOfs);
    else XSetForeground(GRAPHICS_DISPLAY,GRAPHICS_CONTEXT,FIND_GW(window)->foreground+ColorOfs);
  XFillPolygon(GRAPHICS_DISPLAY,window,GRAPHICS_CONTEXT,points,num,0,CoordModeOrigin);
//  XFlush(GRAPHICS_DISPLAY);                                     // no one knows why == 0
}

void fillpoly(int num,polypoint *points,int color) { 
  if (default_window!=MAXINT) 
    fillpoly(default_window,num,points,color); 
  else
    Error("Unkown default window.");
}

void rectangle(Window window,int xa,int ya,int xb,int yb,int color) {
  line(window,xa,ya,xb,ya,color);
  line(window,xb,ya,xb,yb,color);
  line(window,xb,yb,xa,yb,color);
  line(window,xa,yb,xa,ya,color);
}

void rectangle(int xa,int ya,int xb,int yb,int color) { 
  if (default_window!=MAXINT) 
    rectangle(default_window,xa,ya,xb,yb,color);
  else
    Error("Unkown default window.");
}

void bar(Window window,int xa,int ya,int xb,int yb,int color) {
  if (color>-MAXINT) XSetForeground(GRAPHICS_DISPLAY,GRAPHICS_CONTEXT,color+ColorOfs);
    else XSetForeground(GRAPHICS_DISPLAY,GRAPHICS_CONTEXT,FIND_GW(window)->foreground+ColorOfs);
  XFillRectangle(GRAPHICS_DISPLAY,window,GRAPHICS_CONTEXT,xa,ya,xb-xa,yb-ya);
}

void bar(int xa,int ya,int xb,int yb,int color) { 
  if (default_window!=MAXINT) 
    bar(default_window,xa,ya,xb,yb,color);
  else
    Error("Unkown default window.");
}

void putrgbpixel(Window window,int x,int y,int r,int g,int b) {
  XSetForeground(GRAPHICS_DISPLAY,GRAPHICS_CONTEXT,ConvRGBtoPIX(r,g,b));
  XDrawPoint(GRAPHICS_DISPLAY,window,GRAPHICS_CONTEXT,x,y);
//  XFlush(GRAPHICS_DISPLAY);
}

void putrgbpixel(int x,int y,int r,int g,int b) { 
  if (default_window!=MAXINT) 
    putrgbpixel(default_window,x,y,r,g,b);
  else
    Error("Unkown default window.");
}

//-------------------------------------------------------------------

unsigned long imagesize(Window window,int x1,int y1,int x2,int y2) {
  unsigned long s;
  XImage *p;  

  int w=x2-x1+1;
  int h=y2-y1+1;
  p=XGetImage(GRAPHICS_DISPLAY,window,x1,y1,w,h,0xffffffff,ZPixmap);

  s=(unsigned long)(p->bytes_per_line)*(unsigned long)h+6;
  return s;
}

void getimage(Window window,int x,int y,unsigned short w,unsigned short h,void *bitmap) {
  unsigned short *d;
  char *data;
  XImage *p;  
  unsigned long i,j,o;

  data=(char*)bitmap;
  data+=6;
  ((short*)bitmap)[0]=w;
  ((short*)bitmap)[1]=h;

  p=XGetImage(GRAPHICS_DISPLAY,window,x,y,w,h,0xffffffff,ZPixmap);
  o=p->bytes_per_line;
  ((short*)bitmap)[2]=o;
  for (i=0;i<h;i++) for (j=0;j<o;j++) data[i*w+j]=p->data[i*o+j];
}

void getimage(int x,int y,int w,int h,void *bitmap) { 
  if (default_window!=MAXINT) 
    getimage(default_window,x,y,w,h,bitmap);
  else
    Error("Unkown default window.");
}

void putimage(Window window,int x,int y,void *bitmap,int /*flag*/ ) {
  XImage *p;
  unsigned short w,h,*d;
  char *data;

  data=(char*)bitmap;
  d=(unsigned short*)bitmap;
  data+=6;
  w=((short*)bitmap)[0];
  h=((short*)bitmap)[1];
  p=XCreateImage(GRAPHICS_DISPLAY,VisualInfo,8,ZPixmap,0,data,w,h,8,w);
  XPutImage(GRAPHICS_DISPLAY,window,GRAPHICS_CONTEXT,p,0,0,x,y,w,h);
//  XFlush(GRAPHICS_DISPLAY);
}

void putimage(int x,int y,void *bitmap,int flag) { 
  if (default_window!=MAXINT) 
    putimage(default_window,x,y,bitmap,flag);
  else
    Error("Unkown default window.");
}

//-----------------------------------------------------------------------

void outtextxy(Window window,int x,int y,char *s,int color) {
  XTextItem text;

  if (color>-MAXINT) 
    XSetForeground(GRAPHICS_DISPLAY,GRAPHICS_CONTEXT,color+ColorOfs);
  else 
    XSetForeground(GRAPHICS_DISPLAY,GRAPHICS_CONTEXT,FIND_GW(window)->foreground+ColorOfs);

  text.chars=s;
  text.nchars=strlen(s);
  text.delta=1;
  text.font=font;
  XDrawText(GRAPHICS_DISPLAY,window,GRAPHICS_CONTEXT,x,y,&text,1);

}

void outtextxy(int x,int y,char *s,int color) { 
  if (default_window!=MAXINT) 
    outtextxy(default_window,x,y,s,color); 
  else
    Error("Unkown default window.");
}

void outtext(Window window,char *s,int color) {
  GW* p;

  p=FIND_GW(window);
  if (p!=NULL) outtextxy(window,p->penx,p->peny,s,color);
}


void outtext(char *s,int color) { 
  if (default_window!=MAXINT) 
    outtext(default_window,s,color); 
  else
    Error("Unkown default window.");
}

//-----------------------------------------------------------------------

void arc(Window window,int x,int y,int startangle,int stopangle,int r,int color) {
  if (color>-MAXINT) 
    XSetForeground(GRAPHICS_DISPLAY,GRAPHICS_CONTEXT,color+ColorOfs);
  else 
    XSetForeground(GRAPHICS_DISPLAY,GRAPHICS_CONTEXT,FIND_GW(window)->foreground+ColorOfs);
  XDrawArc(GRAPHICS_DISPLAY,window,GRAPHICS_CONTEXT,x-r,y-r,2*r,2*r,startangle/64,stopangle/64);
}

void arc(int x,int y,int startangle,int stopangle,int r,int color) { 
  if (default_window!=MAXINT) 
    arc(default_window,x,y,startangle,stopangle,r,color);
  else
    Error("Unkown default window.");
}

//-----------------------------------------------------------------------

void circle(Window window,int x,int y,int r,int color) { arc(window,x,y,r,0,360,color); }

void circle(int x,int y,int r,int color) { 
  if (default_window!=MAXINT) 
    circle(default_window,x,y,r,color);
  else
    Error("Unkown default window.");
}

//-----------------------------------------------------------------------

void setfont(char *fcode) {

  sprintf(fontname,"%s",fcode);
  font=XLoadFont(GRAPHICS_DISPLAY,fcode);
  XFontStruct *f=XQueryFont(GRAPHICS_DISPLAY,font);
  fontwidth =f->max_bounds.width;
  fontheight=f->max_bounds.ascent+f->max_bounds.descent;
}

char *getfont() { return fontname; }
int getfontwidth() { return fontwidth+1; }
int getfontheight() { return fontheight; }

unsigned int readkey() {
  return EVENTKEYCODE;
}

void xFlush() {
  XFlush(GRAPHICS_DISPLAY);
}


