/***********************************************************************/
/* File:                                                               */
/*                                                                     */
/*                                                                     */
/*---------------------------------------------------------------------*/
/* Programmer: Nils Schmeisser                                         */
/*                                                                     */
/*---------------------------------------------------------------------*/
/* Revision date: 2.12.95                                              */
/*---------------------------------------------------------------------*/
/* Notes:                                                              */
/*                                                                     */
/*                                                                     */
/***********************************************************************/

#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include <xapp.h>
#include <Xm/FileSB.h>

char* Copyright="xpgp V 1.0, (C) Nils Schmeisser, 1997\nThis file ist restricted by the GNU General Public License Agreement.\n";

typedef enum { SIGN, ENCRYPT, SIGN_ENCRYPT, CHECK } PGPMODE;

Widget topshell;

Widget efileselector;
Widget Editor;
Widget st,mb;
char *fname;

PGPMODE pgpmode;

char logfile[1024];
char mailpath[1024];

char passphrase[256];
char receipient[256];

Widget passw,passdlg,receipdlg,receipw;
Widget maildlg,mailw,ccw,subjectw;
Widget xpgphelp,pgphelp,about;
static Widget signdlg;

/*-------------------------------------------------------------------*/

void EditorAddButton(char *name,XtCallbackProc proc) {
  xCreatePushButton(name,kbNO,(XtCallbackProc)proc,mb);
}

char *GetEditorFName() {
  return fname;
}

void OpenEditor(char *fname_) {
  char *s;
  FILE *f;
  struct stat status;

  if (fname_!=NULL) {
    sprintf(fname,"%s",fname_);
    if ((f=fopen(fname,"r"))!=NULL) {
      stat(fname,&status);
      long FileSize = status.st_size;
      s=new char [FileSize+1];
      fread(s,FileSize,1,f);
      s[FileSize]=0;
      XmTextSetString(st,s);
      delete s;
      fclose(f);
    } else
      sprintf(fname,"untitled");
  } else {
    sprintf(fname,"untitled");
    XmTextSetString(st,"");
  }
  
  xPopup(Editor);
}

void store(char *fname) {
  FILE *f;
  char *s;

  s=XmTextGetString(st);
  f=fopen(fname,"w");
  fwrite(s,strlen(s),1,f);
  fclose(f);
  XtFree(s);
}

void load(char *fname) {
  FILE *f;
  char *s;
  struct stat status;

  if ((f=fopen(fname,"r"))!=NULL) {
    stat(fname,&status);
    long FileSize = status.st_size;
    s=new char [FileSize+1];
    fread(s,FileSize,1,f);
    s[FileSize]=0;
    XmTextSetString(st,s);
    delete s;
    fclose(f);
  }
}

void open(Widget,caddr_t,caddr_t out_data) {
  XmFileSelectionBoxCallbackStruct *info;
  FILE *f;
  char *s;
  struct stat status;

  info=(XmFileSelectionBoxCallbackStruct*)out_data;
  XmStringGetLtoR(info->value,XmSTRING_DEFAULT_CHARSET,&fname);
  XtUnmanageChild(efileselector);

  load(fname);
}

void eloadCB(Widget,caddr_t,caddr_t) {
  XtRemoveAllCallbacks(efileselector,XmNokCallback);
  XtAddCallback(efileselector,XmNokCallback,(XtCallbackProc)open,"open");
  XtManageChild(efileselector);
}

void save(Widget,caddr_t,caddr_t out_data) {
  XmFileSelectionBoxCallbackStruct *info;
  FILE *f;
  char *s;

  if (out_data!=NULL) {
    info=(XmFileSelectionBoxCallbackStruct*)out_data;
    XmStringGetLtoR(info->value,XmSTRING_DEFAULT_CHARSET,&fname);
    XtUnmanageChild(efileselector);
  }
  
  store(fname);
}

void esaveasCB(Widget,caddr_t,caddr_t) {
  XtRemoveAllCallbacks(efileselector,XmNokCallback);
  XtAddCallback(efileselector,XmNokCallback,(XtCallbackProc)save,"Save");
  XtManageChild(efileselector);
}

void esaveCB(Widget w,caddr_t a,caddr_t b) {
  if (strlen(fname)==0) 
    esaveasCB(w,a,b);
  else
    save(w,a,NULL);
}

void enewCB(Widget,caddr_t,caddr_t) {
  XmTextSetString(st,"");
}

void eexitCB(Widget,caddr_t,caddr_t) {
  char cmd[1024];
  xPopdown(Editor);
  sprintf(cmd,"rm -f %s",logfile);
  system(cmd);
  exit(0);
}

void ehidefselector(Widget,caddr_t,caddr_t) {
  XtUnmanageChild(efileselector);
}

void xpgpstat() {
  FILE *f;
  char s[1024];
  struct stat fstatus;
  int FileSize;
  int i;
  char cmd[1024];

  s[0]=0;
  if ((f=fopen(logfile,"r"))!=NULL) {
    stat(logfile,&fstatus);
    FileSize = (int)(fstatus.st_size);
    fread(s,FileSize,1,f);
    fclose(f);
    s[FileSize-1]=0;
    for (i=0;i<strlen(s);i++)
      if (((s[i]!='\n') && (s[i]<' ')) || (s[i]>0x7e))
	s[i]=32;
    sprintf(cmd,"grep Guide %s > /dev/null",logfile);
    if (!(system(cmd))) {
      s[strlen(s)-90]=0;
    }
    xCreateMessageBox("XPGP Status",s,topshell);
  }
}

/*-------------------------------------------------------------------*/

void sign(char *fname,char *pphrase) {
  char cmd[1024];

  sprintf(cmd,"rm -f %s.asc ; pgp +batchmode -z%c%s%c -sa %s > %s",
	  fname,0x22,pphrase,0x22,fname,logfile);
  if (!system(cmd)) {
    sprintf(fname,"%s.asc",fname);
    load(fname);
  }
  xpgpstat();
}

void encrypt(char *fname,char *receipient) {
  char cmd[1024];

  sprintf(cmd,"rm -f %s.asc ; pgp +batchmode -ea %s %s > %s",
	  fname,fname,receipient,logfile);
  if (!system(cmd)) {
    sprintf(fname,"%s.asc",fname);
    load(fname);
  }
  xpgpstat();
}

void sign_encrypt(char *fname,char *receipient,char *passphrase) {
  char cmd[1024];

  sprintf(cmd,"rm -f %s.asc ; pgp +batchmode -z%c%s%c -sea %s %s > %s",
	  fname,0x22,passphrase,0x22,fname,receipient,logfile);
  if (!system(cmd)) {
    sprintf(fname,"%s.asc",fname);
    load(fname);
  }
  xpgpstat();
}

void decryptnamed(char *fname,char *passphrase) {
  char cmd[1024];
  char nfname[1024];
  int status,error;

  strcpy(nfname,fname);
  nfname[strlen(fname)-4]=0;
  
  sprintf(cmd,"rm -f %s ; pgp +batchmode -z%c%s%c %s > %s",
	  nfname,0x22,passphrase,0x22,fname,logfile);
  status=system(cmd);
  printf("status:%x\n",status);
  if (!(status & 0x400)) {
    strcpy(fname,nfname);
    load(fname);
  }
  xpgpstat();
}

void decrypt(char *fname) {
  char cmd[1024];
  char nfname[1024];
  int status,error;

  strcpy(nfname,fname);
  nfname[strlen(fname)-4]=0;
  
  sprintf(cmd,"rm -f %s ; pgp +batchmode %s > %s",nfname,fname,logfile);
  status=system(cmd);

  printf("status:%x\n",status);
  if (status & 0x400) {
    sprintf(cmd,"grep Error %s > /dev/null",logfile);
    error=system(cmd);
    if (!error) {
      pgpmode=CHECK;
      xPopup(passdlg);
    } else
      xpgpstat();
  } else {
    strcpy(fname,nfname);
    load(fname);
    xpgpstat();
  }
}

/*-------------------------------------------------------------------*/

void ReceipDlgOKCB(Widget,caddr_t,caddr_t) {
  strcpy(receipient,xGetInputLineString(receipw));
  xSetInputLineValue(receipw,"");
  xPopdown(receipdlg);

  if (pgpmode==ENCRYPT)
    encrypt(fname,receipient);
  if (pgpmode==SIGN_ENCRYPT)
    sign_encrypt(fname,receipient,passphrase);
}

void ReceipDlgCancelCB(Widget,caddr_t,caddr_t) {
  xSetInputLineValue(passw,"");
  xPopdown(receipdlg);
}

Widget InitReceipDlg(Widget topshell) {
  Widget receipdlg,rc,rcb;

  receipdlg=xCreatePopupShell("Receipient",20,20,200,100,topshell);
    rc=xCreateRowColumn(XmVERTICAL,receipdlg);
      receipw=xCreateInputLine("Receipient","",100,rc);
      XtAddCallback(receipw,XmNactivateCallback,(XtCallbackProc)ReceipDlgOKCB,"");
      /*      rcb=xCreateRowColumn(XmHORIZONTAL,rc);
        xCreatePushButton("  Ok  ",kbEnter,(XtCallbackProc)ReceipDlgOKCB,rcb);
        xCreatePushButton("Cancel",kbEnter,(XtCallbackProc)ReceipDlgCancelCB,rcb);*/
  return receipdlg;
}

/*-------------------------------------------------------------------*/

void HiddenInputCB(Widget w,caddr_t a,caddr_t b);

void PassDlgOKCB(Widget,caddr_t,caddr_t) {
  XtRemoveCallback(passw,XmNmodifyVerifyCallback,(XtCallbackProc)HiddenInputCB,"");
  xSetInputLineValue(passw,"");
  XtAddCallback(passw,XmNmodifyVerifyCallback,(XtCallbackProc)HiddenInputCB,"");
  xPopdown(passdlg);

  if (pgpmode==CHECK)
    decryptnamed(fname,passphrase);
  else if (pgpmode==SIGN_ENCRYPT)
    xPopup(receipdlg);
  else {
    /* pgpmode IS SIGN */
    sign(fname,passphrase);
  }
}

void PassDlgCancelCB(Widget,caddr_t,caddr_t) {
  XtRemoveCallback(passw,XmNmodifyVerifyCallback,(XtCallbackProc)HiddenInputCB,"");
  xSetInputLineValue(passw,"");
  XtAddCallback(passw,XmNmodifyVerifyCallback,(XtCallbackProc)HiddenInputCB,"");
  xPopdown(passdlg);
}

void HiddenInputCB(Widget w,caddr_t a,caddr_t b) {
  XmTextVerifyCallbackStruct* data=(XmTextVerifyCallbackStruct*)b;

  char *s;

  if ((data->text->ptr)[0]==0) {
    if ((data->text->ptr)[1]==119) {
      PassDlgOKCB(w,a,b);
    } else if ((data->text->ptr)[1]==101) {
      if (strlen(passphrase)>0)
	passphrase[strlen(passphrase)-1]=0;
    }
  } else {
    sprintf(passphrase,"%s%c",passphrase,data->text->ptr[0]);
    s=new char [strlen((char*)((data->text->ptr)+1))+1+2];
    sprintf(s,"%s*",(data->text->ptr)+1);
    if (data->text->ptr!=NULL) XtFree(data->text->ptr);
    data->text->ptr=s;
    data->text->length=strlen(s);
  }
}

Widget InitPassDlg(Widget topshell) {
  Widget passdlg,rc,rcb;

  passdlg=xCreatePopupShell("User Passphrase",20,20,200,100,topshell);
    rc=xCreateRowColumn(XmVERTICAL,passdlg);
      passw=xCreateInputLine("Your pass phrase","",100,rc);
      XtAddCallback(passw,XmNactivateCallback,(XtCallbackProc)PassDlgOKCB,"");
      XtAddCallback(passw,XmNmodifyVerifyCallback,(XtCallbackProc)HiddenInputCB,"");
      /*      rcb=xCreateRowColumn(XmHORIZONTAL,rc);
        xCreatePushButton("  Ok  ",kbEnter,(XtCallbackProc)PassDlgOKCB,rcb);
        xCreatePushButton("Cancel",kbEnter,(XtCallbackProc)PassDlgCancelCB,rcb); */
  return passdlg;
}

/*-------------------------------------------------------------------*/

void eSignCB(Widget,caddr_t,caddr_t) {
  store(fname);
  strcpy(passphrase,"");
  pgpmode=SIGN;
  xPopup(passdlg);
}

/*-------------------------------------------------------------------*/

void eEncryptCB(Widget,caddr_t,caddr_t) {
  store(fname);
  pgpmode=ENCRYPT;
  xPopup(receipdlg);
}

/*-------------------------------------------------------------------*/

void eSignEncryptCB(Widget,caddr_t,caddr_t) {
  store(fname);
  strcpy(passphrase,"");
  pgpmode=SIGN_ENCRYPT;
  xPopup(passdlg);
}

/*-------------------------------------------------------------------*/

void eDecryptCB(Widget,caddr_t,caddr_t) {
  store(fname);
  strcpy(passphrase,"");
  decrypt(fname);
}

/*-------------------------------------------------------------------*/

void MailDlgOKCB(Widget,caddr_t,caddr_t) {
  char to[256],cc[1024],subject[256],cmd[1024];

  strcpy(to,xGetInputLineString(mailw));
  strcpy(cc,xGetInputLineString(ccw));
  strcpy(subject,xGetInputLineString(subjectw));
  xSetInputLineValue(mailw,"");
  xSetInputLineValue(ccw,"");
  xSetInputLineValue(subjectw,"");
  xPopdown(maildlg);

  if (strlen(cc)>0) 
    sprintf(cmd,"mail -s %c%s%c -c %s %s < %s",
	    0x22,subject,0x22,cc,to,fname);
  else
    sprintf(cmd,"mail -s %c%s%c %s < %s",
	    0x22,subject,0x22,to,fname);
  system(cmd);
}

void MailDlgCancelCB(Widget,caddr_t,caddr_t) {
  xSetInputLineValue(mailw,"");
  xSetInputLineValue(ccw,"");
  xSetInputLineValue(subjectw,"");

  xPopdown(maildlg);
}

Widget InitMailDlg(Widget topshell) {
  Widget maildlg,rc,rcb;

  maildlg=xCreatePopupShell("Send mail",20,20,200,100,topshell);
    rc=xCreateRowColumn(XmVERTICAL,maildlg);
      subjectw=xCreateInputLine("Subject","",200,rc);
      mailw   =xCreateInputLine("To","",200,rc);
      ccw     =xCreateInputLine("CC","",200,rc);
      XtAddCallback(mailw,XmNactivateCallback,(XtCallbackProc)MailDlgOKCB,"");
      rcb=xCreateRowColumn(XmHORIZONTAL,rc);
        xCreatePushButton("  Ok  ",kbEnter,(XtCallbackProc)MailDlgOKCB,rcb);
        xCreatePushButton("Cancel",kbEnter,(XtCallbackProc)MailDlgCancelCB,rcb);
  return maildlg;
}

void emailCB(Widget,caddr_t,caddr_t) {
  store(fname);
  xPopup(maildlg);
}

void ereadCB(Widget,caddr_t,caddr_t) {
  struct stat status;
  int fsize;

  stat(mailpath,&status);
  fsize=(int)(status.st_size);
  if (fsize>0) {
    store(fname);
    strcpy(fname,mailpath);
    load(fname);
    sprintf(fname,"untitled.asc");
    store(fname);
  } else
    xCreateMessageBox("Mail","No mail available.\n",topshell);
}

/*-------------------------------------------------------------------*/

void epgpCB(Widget w,caddr_t,caddr_t) {
  char wn[128];

  strcpy(wn,XtName(w));
  printf("%s\n",wn);
}

/*-------------------------------------------------------------------*/

void InitHelp() {
  xpgphelp=xCreateMessageBox("Help on XPGP","",topshell,FALSE); xPopdown(xpgphelp);
  pgphelp =xCreateMessageBox("Help on XPGP","",topshell,FALSE); xPopdown(pgphelp);
  about   =xCreateMessageBox("About",Copyright,topshell,FALSE); xPopdown(about);
}

/*-------------------------------------------------------------------*/

void AboutCB(Widget,caddr_t,caddr_t) {
  xPopup(about);
}

void ehelpxpgpCB(Widget,caddr_t,caddr_t) {
  xPopup(xpgphelp);
}

void ehelppgpCB(Widget,caddr_t,caddr_t) {
  xPopup(pgphelp);
}

/*-------------------------------------------------------------------*/

Widget InitEditor(Widget parent) {
  Arg args[10];
  int n=0,i;
  Widget r,rcb;
  Widget sm,sm2;
  Widget Editor;

  fname=new char[1024];
  strcpy(fname,"untitled");
  efileselector=XmCreateFileSelectionDialog(parent,"File Selection",NULL,0);  
  XtAddCallback(efileselector,XmNcancelCallback,(XtCallbackProc)ehidefselector,"Hide");

  Editor=xCreatePopupShell("XPGP V 1.0",100,100,600,400,parent);
    r=xCreateRowColumn(XmVERTICAL,Editor);

      mb=XmCreateMenuBar(r,"emenubar",(Arg*)NULL,0);
      XtManageChild(mb);

      sm=XmCreatePulldownMenu(mb,"eFile_",(Arg*)NULL,0);
        xCreatePushButton("New",kbNO,(XtCallbackProc)enewCB,sm);
        xCreatePushButton("Open",kbNO,(XtCallbackProc)eloadCB,sm);
        xCreatePushButton("Save",kbNO,(XtCallbackProc)esaveCB,sm);
        xCreatePushButton("Save As ...",kbNO,(XtCallbackProc)esaveasCB,sm);
	xCreateSeparator(sm);
        xCreatePushButton("Exit",kbNO,(XtCallbackProc)eexitCB,sm);
      n=0;
      XtSetArg(args[n],XmNsubMenuId,sm); n++;
      XtCreateManagedWidget("File",xmCascadeButtonWidgetClass,mb,args,n);
    
      sm=XmCreatePulldownMenu(mb,"eCERTIFY_",(Arg*)NULL,0);
        xCreatePushButton("Sign",kbNO,(XtCallbackProc)eSignCB,sm);
        xCreatePushButton("Encrypt",kbNO,(XtCallbackProc)eEncryptCB,sm);
        xCreatePushButton("Sign & Encrypt",kbNO,(XtCallbackProc)eSignEncryptCB,sm);
	xCreateSeparator(sm);
        xCreatePushButton("Check Sign & Decrypt",kbNO,(XtCallbackProc)eDecryptCB,sm);
      n=0;
      XtSetArg(args[n],XmNsubMenuId,sm); n++;
      XtCreateManagedWidget("Certify",xmCascadeButtonWidgetClass,mb,args,n);
    
      sm=XmCreatePulldownMenu(mb,"eMAIL_",(Arg*)NULL,0);
        xCreatePushButton("Send",kbNO,(XtCallbackProc)emailCB,sm);
        xCreateSeparator(sm);
        xCreatePushButton("Read",kbNO,(XtCallbackProc)ereadCB,sm);
      n=0;
      XtSetArg(args[n],XmNsubMenuId,sm); n++;
      XtCreateManagedWidget("Mail",xmCascadeButtonWidgetClass,mb,args,n);
    
      sm=XmCreatePulldownMenu(mb,"ePGP_",(Arg*)NULL,0);
        xCreatePushButton("Generate new key pair",kbNO,(XtCallbackProc)epgpCB,sm);
        xCreatePushButton("Add key from file",kbNO,(XtCallbackProc)epgpCB,sm);
        xCreatePushButton("Extract key",kbNO,(XtCallbackProc)epgpCB,sm);
        xCreatePushButton("View key(s)",kbNO,(XtCallbackProc)epgpCB,sm);
        xCreatePushButton("View fingerprint",kbNO,(XtCallbackProc)epgpCB,sm);
        xCreatePushButton("Check & View in detail",kbNO,(XtCallbackProc)epgpCB,sm);
        xCreatePushButton("Remove userid or key",kbNO,(XtCallbackProc)epgpCB,sm);
        xCreatePushButton("Edit trust params",kbNO,(XtCallbackProc)epgpCB,sm);
        xCreatePushButton("Add another userid",kbNO,(XtCallbackProc)epgpCB,sm);
        xCreatePushButton("Edit passphrase",kbNO,(XtCallbackProc)epgpCB,sm);
        xCreatePushButton("Sign a key in pubring",kbNO,(XtCallbackProc)epgpCB,sm);
        xCreatePushButton("Remove a sign from key",kbNO,(XtCallbackProc)epgpCB,sm);
        xCreatePushButton("Revoke disable/enable",kbNO,(XtCallbackProc)epgpCB,sm);
      n=0;
      XtSetArg(args[n],XmNsubMenuId,sm); n++;
      XtCreateManagedWidget("PGP",xmCascadeButtonWidgetClass,mb,args,n);
    
      sm=XmCreatePulldownMenu(mb,"eHELP_",(Arg*)NULL,0);
        xCreatePushButton("On XPGP",kbNO,(XtCallbackProc)ehelpxpgpCB,sm);
        xCreatePushButton("On PGP",kbNO,(XtCallbackProc)ehelppgpCB,sm);
	xCreateSeparator(sm);
        xCreatePushButton("About XPGP",kbNO,(XtCallbackProc)AboutCB,sm);
      n=0;
      XtSetArg(args[n],XmNsubMenuId,sm); n++;
      XtCreateManagedWidget("Help",xmCascadeButtonWidgetClass,mb,args,n);
    
      n=0;
      XtSetArg(args[n],XmNwidth,400);  n++;
      XtSetArg(args[n],XmNrows,40);  n++;
      XtSetArg(args[n],XmNeditMode,XmMULTI_LINE_EDIT);  n++;
      st=XmCreateScrolledText(r,"",args,n);
      XtManageChild(st);
      XmTextSetString(st,"");
  return Editor;
}

/*-------------------------------------------------------------------*/

int main(int argc,char **argv) {
   XtAppContext ctx;
   Widget menushell;
   Arg args[10];

   printf("%s\n",Copyright);
   sprintf(logfile,"/tmp/xpgp.%d",getuid());
   sprintf(mailpath,getenv("MAIL"));

   //--- Creating applications basic views ------------------------------------------

   topshell=xInitApplication("xpgp V1.0",&ctx,argc,argv);
     passdlg=InitPassDlg(topshell);
     receipdlg=InitReceipDlg(topshell);
     maildlg=InitMailDlg(topshell);
     Editor=InitEditor(topshell);
   
   InitHelp();

   xPopup(Editor);

   XtAppMainLoop(ctx);
   return 0;
}
